@@ -25,7 +25,8 @@
 #include <Python.h>
 #include <structmember.h>
 
-#define MAX_ANTENNA_COUNT 4
+#define MAX_ANTENNA_COUNT 16
+#define MAX_GPIO_COUNT 4
 #define MAX_DATA_AREA 258
 #define numberof(x) (sizeof((x))/sizeof((x)[0]))
 
@@ -33,6 +34,7 @@ typedef struct {
     PyObject_HEAD
     /* Type-specific fields go here. */
     TMR_Reader reader;
+    TMR_TagOp tagop;
     uint8_t antennas[MAX_ANTENNA_COUNT];
     TMR_ReadListenerBlock readListener;
     PyObject *readCallback;
@@ -321,7 +323,7 @@ Reader_set_read_plan(Reader *self, PyObject *args, PyObject *kwds)
         return NULL;
     }
 
-    if ((ant_count = PyList_Size(list)) > 4)
+    if ((ant_count = PyList_Size(list)) > MAX_ANTENNA_COUNT)
     {
         PyErr_SetString(PyExc_TypeError, "Too many antennas");
         return NULL;
@@ -345,7 +347,6 @@ Reader_set_read_plan(Reader *self, PyObject *args, PyObject *kwds)
     if (bank != NULL)
     {
         int op = 0;
-        TMR_TagOp tagop;
 
         if(PyList_Check(bank))
         {
@@ -365,10 +366,10 @@ Reader_set_read_plan(Reader *self, PyObject *args, PyObject *kwds)
                 return NULL;
         }
 
-        if ((ret = TMR_TagOp_init_GEN2_ReadData(&tagop, op, 0, 0)) != TMR_SUCCESS)
+        if ((ret = TMR_TagOp_init_GEN2_ReadData(&self->tagop, op, 0, 0)) != TMR_SUCCESS)
             goto fail;
 
-        if ((ret = TMR_RP_set_tagop(&plan, &tagop)) != TMR_SUCCESS)
+        if ((ret = TMR_RP_set_tagop(&plan, &self->tagop)) != TMR_SUCCESS)
             goto fail;
     }
 
@@ -561,6 +562,178 @@ fail:
     return NULL;
 }
 
+static PyObject *
+Reader_get_write_powers(Reader *self)
+{
+    int row;
+    TMR_Status ret;
+    PyObject *antenna_power;
+    PyObject *antenna_powers;
+    antenna_powers = PyList_New(0);
+    TMR_PortValueList ant_pow_list;
+    TMR_PortValue pow_value_list[MAX_ANTENNA_COUNT];
+
+    ant_pow_list.list = pow_value_list;
+    ant_pow_list.max = numberof(pow_value_list);
+
+    TMR_uint8List port_list;
+    uint8_t port_value_list[MAX_ANTENNA_COUNT];
+
+    port_list.list = port_value_list;
+    port_list.max = numberof(port_value_list);
+
+    if ((ret = TMR_paramGet(&self->reader, TMR_PARAM_ANTENNA_PORTLIST, &port_list)) != TMR_SUCCESS)
+    {
+        PyErr_SetString(PyExc_TypeError, "Error getting antennas");
+        return NULL;
+    }
+
+    if ((ret = TMR_paramGet(&self->reader, TMR_PARAM_RADIO_PORTWRITEPOWERLIST, &ant_pow_list)) != TMR_SUCCESS)
+    {
+        PyErr_SetString(PyExc_TypeError, TMR_strerr(&self->reader, ret));
+        return NULL;
+    }
+
+    for (row = 0; row < port_list.len; row++)
+    {
+        antenna_power = PyTuple_New(2);
+        PyTuple_SetItem(antenna_power, 0, PyLong_FromLong((long) ant_pow_list.list[row].port));
+        PyTuple_SetItem(antenna_power, 1, PyLong_FromLong((long) ant_pow_list.list[row].value));
+        PyList_Append(antenna_powers, antenna_power);
+    }
+
+    return antenna_powers;
+}
+
+static PyObject *
+Reader_set_write_powers(Reader *self, PyObject *args, PyObject *kwds)
+{
+    int length;
+    TMR_Status ret;
+    uint8_t ant_count, pow_count;
+    PyObject *power_list, *antenna_list;
+    static char *kwlist[] = {"antennas", "powers", NULL};
+
+    if (!PyArg_ParseTupleAndKeywords(args, kwds, "O!O!", kwlist, &PyList_Type, &antenna_list, &PyList_Type, &power_list))
+        return NULL;
+    if ((ant_count = PyList_Size(antenna_list)) > MAX_ANTENNA_COUNT)
+    {
+        PyErr_SetString(PyExc_TypeError, "Too many antennas");
+        return NULL;
+    }
+    if ((pow_count = PyList_Size(power_list)) > MAX_ANTENNA_COUNT)
+    {
+        PyErr_SetString(PyExc_TypeError, "Too many powers");
+        return NULL;
+    }
+    if (pow_count != ant_count)
+    {
+        PyErr_SetString(PyExc_TypeError, "Number of antennas and powers not matching");
+        return NULL;
+    }
+    length = (int) ant_count;
+
+    int row;
+    int power;
+    int antenna;
+    TMR_PortValueList ant_pow_list;
+    TMR_PortValue value_list[MAX_ANTENNA_COUNT];
+
+    ant_pow_list.len = length;
+    ant_pow_list.max = numberof(value_list);
+    ant_pow_list.list = value_list;
+    for (row = 0; row < length; row++)
+    {
+        power = (int) PyLong_AsLong(PyList_GetItem(power_list, row));
+        antenna = (int) PyLong_AsLong(PyList_GetItem(antenna_list, row));
+
+        if ((ant_pow_list.list[row].port = antenna) == 255)
+        {
+            PyErr_SetString(PyExc_TypeError, "antennas expecting a list of integers");
+            return NULL;
+        }
+        if ((ant_pow_list.list[row].value = power) == 255)
+        {
+            PyErr_SetString(PyExc_TypeError, "powers expecting a list of integers");
+            return NULL;
+        }
+    }
+
+    if (length > 0)
+    {
+        if ((ret = TMR_paramSet(&self->reader, TMR_PARAM_RADIO_PORTWRITEPOWERLIST, &ant_pow_list)) != TMR_SUCCESS)
+            goto fail;
+    }
+
+    return Reader_get_write_powers(self);
+fail:
+    PyErr_SetString(PyExc_TypeError, TMR_strerr(&self->reader, ret));
+    return NULL;
+}
+
+static PyObject *
+Reader_get_antenna_portswitchgpos(Reader *self)
+{
+    TMR_Status ret;
+    PyObject *gpos;
+    TMR_uint8List gpo_list;
+    uint8_t value_list[MAX_GPIO_COUNT];
+    uint8_t i;
+
+    gpo_list.list = value_list;
+    gpo_list.max = numberof(value_list);
+
+    if ((ret = TMR_paramGet(&self->reader, TMR_PARAM_ANTENNA_PORTSWITCHGPOS, &gpo_list)) != TMR_SUCCESS)
+    {
+        PyErr_SetString(PyExc_TypeError, "Error getting gpos");
+        return NULL;
+    }
+
+    gpos = PyList_New(0);
+    for (i = 0; i < gpo_list.len && i < gpo_list.max; i++)
+    {
+        PyList_Append(gpos, PyLong_FromLong((long) gpo_list.list[i]));
+    }
+
+    return gpos;
+}
+
+static PyObject *
+Reader_set_antenna_portswitchgpos(Reader *self, PyObject *args)
+{
+    TMR_Status ret;
+    PyObject *gpos;
+    TMR_uint8List gpo_list;
+    uint8_t value_list[MAX_GPIO_COUNT];
+    uint8_t gpo_count;
+    uint8_t i;
+
+    if (!PyArg_ParseTuple(args, "O", &gpos))
+        return NULL;
+    if ((gpo_count = PyList_Size(gpos)) > MAX_GPIO_COUNT)
+    {
+        PyErr_SetString(PyExc_TypeError, "Too many gpos");
+        return NULL;
+    }
+
+    gpo_list.len = gpo_count;
+    gpo_list.list = value_list;
+    gpo_list.max = numberof(value_list);
+
+    for (i = 0; i < gpo_list.len && i < gpo_list.max; i++)
+    {
+        value_list[i] = (uint8_t) PyLong_AsLong(PyList_GetItem(gpos, i));
+    }
+
+    if ((ret = TMR_paramSet(&self->reader, TMR_PARAM_ANTENNA_PORTSWITCHGPOS, &gpo_list)) != TMR_SUCCESS)
+    {
+        PyErr_SetString(PyExc_TypeError, TMR_strerr(&self->reader, ret));
+        return NULL;
+    }
+
+    Py_RETURN_NONE; 
+}
+
 static PyObject *
 Reader_write(Reader *self, PyObject *args, PyObject *kwds)
 {
@@ -611,7 +784,10 @@ Reader_read(Reader *self, PyObject *args, PyObject *kwds)
     if (!PyArg_ParseTupleAndKeywords(args, kwds, "|i", kwlist, &timeout))
         return NULL;
 
+    Py_BEGIN_ALLOW_THREADS
     ret = TMR_read(&self->reader, timeout, NULL);
+    Py_END_ALLOW_THREADS
+
     /* In case of TAG ID Buffer Full, extract the tags present in buffer. */
     if (ret != TMR_SUCCESS && ret != TMR_ERROR_TAG_ID_BUFFER_FULL)
     {
@@ -770,6 +946,214 @@ Reader_get_model(Reader* self)
     return PyUnicode_FromString(model.value);
 }
 
+static PyObject *
+Reader_get_gen2_blf(Reader* self)
+{
+    TMR_Status ret;
+    TMR_GEN2_LinkFrequency blf_val;
+
+    if ((ret = TMR_paramGet(&self->reader, TMR_PARAM_GEN2_BLF, &blf_val)) != TMR_SUCCESS)
+    {
+        PyErr_SetString(PyExc_TypeError, TMR_strerr(&self->reader, ret));
+        return NULL;
+    }
+
+    return PyLong_FromLong(blf_val);
+}
+
+static PyObject *
+Reader_set_gen2_blf(Reader* self, PyObject *args)
+{
+    TMR_Status ret;
+    TMR_GEN2_LinkFrequency blf_val;
+
+    if (!PyArg_ParseTuple(args, "i", &blf_val))
+        return NULL;
+
+    if ((ret = TMR_paramSet(&self->reader, TMR_PARAM_GEN2_BLF, &blf_val)) != TMR_SUCCESS)
+    {
+        PyErr_SetString(PyExc_TypeError, TMR_strerr(&self->reader, ret));
+        return NULL;
+    }
+
+    return PyLong_FromLong(blf_val);
+}
+
+static PyObject *
+Reader_get_gen2_tari(Reader* self)
+{
+    TMR_Status ret;
+    TMR_GEN2_Tari tari_val;
+
+    if ((ret = TMR_paramGet(&self->reader, TMR_PARAM_GEN2_TARI, &tari_val)) != TMR_SUCCESS)
+    {
+        PyErr_SetString(PyExc_TypeError, TMR_strerr(&self->reader, ret));
+        return NULL;
+    }
+
+    return PyLong_FromLong(tari_val);
+}
+
+static PyObject *
+Reader_set_gen2_tari(Reader* self, PyObject *args)
+{
+    TMR_Status ret;
+    TMR_GEN2_Tari tari_val;
+
+    if (!PyArg_ParseTuple(args, "i", &tari_val))
+        return NULL;
+
+    if ((ret = TMR_paramSet(&self->reader, TMR_PARAM_GEN2_TARI, &tari_val)) != TMR_SUCCESS)
+    {
+        PyErr_SetString(PyExc_TypeError, TMR_strerr(&self->reader, ret));
+        return NULL;
+    }
+
+    return PyLong_FromLong(tari_val);
+}
+
+static PyObject *
+Reader_get_gen2_tagencoding(Reader* self)
+{
+    TMR_Status ret;
+    TMR_GEN2_TagEncoding tagencoding_val;
+
+    if ((ret = TMR_paramGet(&self->reader, TMR_PARAM_GEN2_TAGENCODING, &tagencoding_val)) != TMR_SUCCESS)
+    {
+        PyErr_SetString(PyExc_TypeError, TMR_strerr(&self->reader, ret));
+        return NULL;
+    }
+
+    return PyLong_FromLong(tagencoding_val);
+}
+
+static PyObject *
+Reader_set_gen2_tagencoding(Reader* self, PyObject *args)
+{
+    TMR_Status ret;
+    TMR_GEN2_TagEncoding tagencoding_val;
+
+    if (!PyArg_ParseTuple(args, "i", &tagencoding_val))
+        return NULL;
+
+    if ((ret = TMR_paramSet(&self->reader, TMR_PARAM_GEN2_TAGENCODING, &tagencoding_val)) != TMR_SUCCESS)
+    {
+        PyErr_SetString(PyExc_TypeError, TMR_strerr(&self->reader, ret));
+        return NULL;
+    }
+
+    return PyLong_FromLong(tagencoding_val);
+}
+
+static PyObject *
+Reader_get_gen2_session(Reader* self)
+{
+    TMR_Status ret;
+    TMR_GEN2_Session session_val;
+
+    if ((ret = TMR_paramGet(&self->reader, TMR_PARAM_GEN2_SESSION, &session_val)) != TMR_SUCCESS)
+    {
+        PyErr_SetString(PyExc_TypeError, TMR_strerr(&self->reader, ret));
+        return NULL;
+    }
+
+    return PyLong_FromLong(session_val);
+}
+
+static PyObject *
+Reader_set_gen2_session(Reader* self, PyObject *args)
+{
+    TMR_Status ret;
+    TMR_GEN2_Session session_val;
+
+    if (!PyArg_ParseTuple(args, "i", &session_val))
+        return NULL;
+
+    if ((ret = TMR_paramSet(&self->reader, TMR_PARAM_GEN2_SESSION, &session_val)) != TMR_SUCCESS)
+    {
+        PyErr_SetString(PyExc_TypeError, TMR_strerr(&self->reader, ret));
+        return NULL;
+    }
+
+    return PyLong_FromLong(session_val);
+}
+
+static PyObject *
+Reader_get_gen2_target(Reader* self)
+{
+    TMR_Status ret;
+    TMR_GEN2_Target target_val;
+
+    if ((ret = TMR_paramGet(&self->reader, TMR_PARAM_GEN2_TARGET, &target_val)) != TMR_SUCCESS)
+    {
+        PyErr_SetString(PyExc_TypeError, TMR_strerr(&self->reader, ret));
+        return NULL;
+    }
+
+    return PyLong_FromLong(target_val);
+}
+
+static PyObject *
+Reader_set_gen2_target(Reader* self, PyObject *args)
+{
+    TMR_Status ret;
+    TMR_GEN2_Target target_val;
+
+    if (!PyArg_ParseTuple(args, "i", &target_val))
+        return NULL;
+
+    if ((ret = TMR_paramSet(&self->reader, TMR_PARAM_GEN2_TARGET, &target_val)) != TMR_SUCCESS)
+    {
+        PyErr_SetString(PyExc_TypeError, TMR_strerr(&self->reader, ret));
+        return NULL;
+    }
+
+    return PyLong_FromLong(target_val);
+}
+
+static PyObject *
+Reader_get_gen2_q(Reader* self)
+{
+    TMR_Status ret;
+    TMR_SR_GEN2_Q model;
+    PyObject *q_value;
+
+    if ((ret = TMR_paramGet(&self->reader, TMR_PARAM_GEN2_Q, &model)) != TMR_SUCCESS)
+    {
+        PyErr_SetString(PyExc_TypeError, TMR_strerr(&self->reader, ret));
+        return NULL;
+    }
+
+    q_value = PyTuple_New(2);
+    PyTuple_SetItem(q_value, 0, PyLong_FromLong((long) model.type));
+    PyTuple_SetItem(q_value, 1, PyLong_FromLong((long) model.u.staticQ.initialQ));
+
+    return q_value;
+}
+
+static PyObject *
+Reader_set_gen2_q(Reader* self, PyObject *args)
+{
+    TMR_Status ret;
+    TMR_SR_GEN2_Q model;
+    PyObject *q_value;
+
+    if (!PyArg_ParseTuple(args, "ii", &model.type, &model.u.staticQ.initialQ))
+        return NULL;
+
+    if ((ret = TMR_paramSet(&self->reader, TMR_PARAM_GEN2_Q, &model)) != TMR_SUCCESS)
+    {
+        PyErr_SetString(PyExc_TypeError, TMR_strerr(&self->reader, ret));
+        return NULL;
+    }
+
+    q_value = PyTuple_New(2);
+    PyTuple_SetItem(q_value, 0, PyLong_FromLong((long) model.type));
+    PyTuple_SetItem(q_value, 1, PyLong_FromLong((long) model.u.staticQ.initialQ));
+
+    return q_value;
+}
+
 static PyMethodDef Reader_methods[] = {
     {"get_temperature", (PyCFunction)Reader_get_temperature, METH_NOARGS,
      "Returns the chip temperature"
@@ -783,6 +1167,9 @@ static PyMethodDef Reader_methods[] = {
     {"get_read_powers", (PyCFunction)Reader_get_read_powers, METH_NOARGS,
      "Lists configured read powers for each antenna."
     },
+    {"get_write_powers", (PyCFunction)Reader_get_write_powers, METH_NOARGS,
+     "Lists configured write powers for each antenna."
+    },
     {"get_supported_regions", (PyCFunction)Reader_get_supported_regions, METH_NOARGS,
      "Returns a list of regions supported by the reader"
     },
@@ -795,6 +1182,15 @@ static PyMethodDef Reader_methods[] = {
     {"set_read_powers", (PyCFunction)Reader_set_read_powers, METH_VARARGS | METH_KEYWORDS,
      "Set the read power for each listed antenna and return the real setted values."
     },
+    {"set_write_powers", (PyCFunction)Reader_set_write_powers, METH_VARARGS | METH_KEYWORDS,
+     "Set the write power for each listed antenna and return the real setted values."
+    },
+    {"get_antenna_portswitchgpos", (PyCFunction)Reader_get_antenna_portswitchgpos, METH_NOARGS,
+     "Lists the GPO pins used for antenna port switching."
+    },
+    {"set_antenna_portswitchgpos", (PyCFunction)Reader_set_antenna_portswitchgpos, METH_VARARGS,
+     "Set the GPO pins used for antenna port switching."
+    },
     {"write", (PyCFunction)Reader_write, METH_VARARGS | METH_KEYWORDS,
      "Write the epc_target tag with the given epc_code"
     },
@@ -810,6 +1206,42 @@ static PyMethodDef Reader_methods[] = {
     {"get_model", (PyCFunction)Reader_get_model, METH_NOARGS,
      "Returns the model name"
     },
+    {"get_gen2_blf", (PyCFunction)Reader_get_gen2_blf, METH_NOARGS,
+     "Returns the current Gen2 BLF setting"
+    },
+    {"set_gen2_blf", (PyCFunction)Reader_set_gen2_blf, METH_VARARGS,
+     "Sets the Gen2 BLF"
+    },
+    {"get_gen2_tari", (PyCFunction)Reader_get_gen2_tari, METH_NOARGS,
+     "Returns the current Gen2 Tari setting"
+    },
+    {"set_gen2_tari", (PyCFunction)Reader_set_gen2_tari, METH_VARARGS,
+     "Sets the Gen2 Tari"
+    },
+    {"get_gen2_tagencoding", (PyCFunction)Reader_get_gen2_tagencoding, METH_NOARGS,
+     "Returns the current Gen2 TagEncoding setting"
+    },
+    {"set_gen2_tagencoding", (PyCFunction)Reader_set_gen2_tagencoding, METH_VARARGS,
+     "Sets the Gen2 TagEncoding"
+    },
+    {"get_gen2_session", (PyCFunction)Reader_get_gen2_session, METH_NOARGS,
+     "Returns the current Gen2 Session setting"
+    },
+    {"set_gen2_session", (PyCFunction)Reader_set_gen2_session, METH_VARARGS,
+     "Sets the Gen2 Session"
+    },
+    {"get_gen2_target", (PyCFunction)Reader_get_gen2_target, METH_NOARGS,
+     "Returns the current Gen2 Target setting"
+    },
+    {"set_gen2_target", (PyCFunction)Reader_set_gen2_target, METH_VARARGS,
+     "Sets the Gen2 Target"
+    },
+    {"get_gen2_q", (PyCFunction)Reader_get_gen2_q, METH_NOARGS,
+     "Returns the current Gen2 Q setting"
+    },
+    {"set_gen2_q", (PyCFunction)Reader_set_gen2_q, METH_VARARGS,
+     "Sets the Gen2 Q"
+    },
     {NULL}  /* Sentinel */
 };
 
